/*
 * Copyright (c) 2020 - 2021 Andri Yngvason
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
 * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

#pragma once

#include "neatvnc.h"

#include <pixman.h>

void nvnc_transform_to_pixman_transform(pixman_transform_t* dst,
		enum nvnc_transform src, int width, int height);

void nvnc_transform_dimensions(enum nvnc_transform transform, uint32_t* width,
		uint32_t* height);

void nvnc_transform_region(struct pixman_region16* dst,
		struct pixman_region16* src, enum nvnc_transform transform,
		int width, int height);

bool nvnc_is_transform_90_degrees(enum nvnc_transform transform);
