/*
 * Copyright (c) 2021 - 2025 Andri Yngvason
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
 * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

#pragma once

#include <stdint.h>

struct nvnc_frame;
struct pixman_region16;
struct nvnc_composite_fb;

struct compositor;

typedef void (*compositor_fn)(struct nvnc_composite_fb*,
		struct pixman_region16* damage, void* userdata);

void composite_buffer_now(struct nvnc_frame* dst, struct nvnc_composite_fb* src,
		struct pixman_region16* damage);

struct compositor* compositor_create(void);
void compositor_destroy(struct compositor*);

int compositor_feed(struct compositor*, struct nvnc_composite_fb* fb,
		struct pixman_region16* damage, compositor_fn on_done,
		void* userdata);
